import numpy as np
from typing import Optional, Tuple

import humba.jits as hj


def histogram(
    x: np.ndarray,
    bins: int = 10,
    range: Tuple[float, float] = (0, 10),
    weights: Optional[np.ndarray] = None,
    flow: bool = False,
) -> Tuple[np.ndarray, Optional[np.ndarray], np.ndarray]:
    """Calculate the histogram for the data ``x``.

    Parameters
    ----------
    x : np.ndarray
        data to histogram
    bins : np.ndarray
        number of bins
    range : (float, float)
        axis range
    weights : np.ndarray, optional
        array of weights for ``x``
    flow : bool
        include over and underflow content in first and last bins

    Returns
    -------
    count : np.ndarray
        The values of the histogram
    error : np.ndarray, optional
        The poission uncertainty on the bin heights
    edges : np.ndarray
        The bin edges

    Notes
    -----
    If the dtype of the ``weights`` is not the same as ``x``, then it
    is converted to the dtype of ``x``.

    Examples
    --------
    >>> import numpy as np
    >>> from humba import histogram
    >>> x = np.random.randn(100000)
    >>> w = np.random.uniform(0.4, 0.5, x.shape[0])
    >>> hist1, _, edges = humba.histogram(x, bins=50, range=(-5, 5))
    >>> hist2, _, edges = humba.histogram(x, bins=50, range=(-5, 5), flow=True)
    >>> hist3, error, edges = histogram(x, bins=50, range=(-5, 5), weights=w)
    >>> hist4, error, edges = histogram(x, bins=50, range=(-3, 3), weights=w, flow=True)

    """
    edges = np.linspace(range[0], range[1], bins + 1)
    if weights is not None:
        assert x.shape == weights.shape, "x and weights must have identical shape"
        if x.dtype == np.float64:
            hfunc = hj._float64_weighted
        elif x.dtype == np.float32:
            hfunc = hj._float32_weighted
        else:
            raise TypeError("dtype of input must be float32 or float64")
        res, err = hfunc(x, weights.astype(x.dtype), bins, range[0], range[1], flow)
        return (res, err, edges)
    else:
        if x.dtype == np.float64:
            hfunc = hj._float64
        elif x.dtype == np.float32:
            hfunc = hj._float32
        else:
            raise TypeError("dtype of input must be float32 or float64")
        res = hfunc(x, bins, range[0], range[1], flow)
        return (res, None, edges)


def histogram_mw(
    x: np.ndarray,
    weights: np.ndarray,
    bins: int = 10,
    range: Tuple[float, float] = (0, 10),
    flow: bool = False,
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Histogram the same data but with multiple weight variations.

    Paramters
    ---------
    x : np.ndarray
        data to histogram
    weights : np.ndarray, optional
        multidimensional array of weights for ``x`` the first element
        of the ``shape`` attribute must be equal to the length of ``x``.
    bins : np.ndarray
        number of bins
    range : (float, float)
        axis range
    flow : bool
        include over and underflow content in first and last bins

    Returns
    -------
    count : np.ndarray
        The values of the histograms calculated from the weights
        Shape will be (bins, ``weights.shape[0]``)
    error : np.ndarray
        The poission uncertainty on the bin heights (shape will be
        the same as ``count``.
    edges : np.ndarray
        The bin edges

    Notes
    -----
    If the dtype of the ``weights`` is not the same as ``x``, then it
    is converted to the dtype of ``x``.

    """
    edges = np.linspace(range[0], range[1], bins + 1)
    assert x.shape[0] == weights.shape[0], "weights shape is not compatible with x"
    if x.dtype == np.float64:
        hfunc = hj._float64_multiweights
    elif x.dtype == np.float32:
        hfunc = hj._float32_multiweights
    else:
        raise TypeError("dtype of input must be float32 or float64")
    res, err = hfunc(x, weights.astype(x.dtype), bins, range[0], range[1], flow)
    return (res, err, edges)



import numba as nb
import numpy as np


@nb.jit(
    nb.types.UniTuple(nb.float32[:], 2)(
        nb.float32[:], nb.float32[:], nb.int64, nb.float64, nb.float64, nb.boolean
    ),
    nopython=True,
)
def _float32_weighted(x, weights, nbins, xmin, xmax, flow):
    norm = 1.0 / (xmax - xmin)
    count = np.zeros((nbins + 2), dtype=np.float32)
    sumw2 = np.zeros((nbins + 2), dtype=np.float32)
    for i in range(x.shape[0]):
        binid = 0
        if x[i] < xmin:
            pass
        elif x[i] > xmax:
            binid = nbins + 1
        else:
            binid = int((x[i] - xmin) * nbins * norm) + 1
        weight = weights[i]
        count[binid] += weight
        sumw2[binid] += weight ** 2
    if flow:
        count[-2] += count[-1]
        sumw2[-2] += sumw2[-1]
        count[1] += count[0]
        sumw2[1] += sumw2[0]
    count, sumw2 = count[1:-1], sumw2[1:-1]
    return count, np.sqrt(sumw2)


@nb.jit(
    nb.float32[:](nb.float32[:], nb.int64, nb.float64, nb.float64, nb.boolean),
    nopython=True,
)
def _float32(x, nbins, xmin, xmax, flow):
    norm = 1.0 / (xmax - xmin)
    count = np.zeros((nbins + 2), dtype=np.float32)
    for i in range(x.shape[0]):
        binid = 0
        if x[i] < xmin:
            pass
        elif x[i] > xmax:
            binid = nbins + 1
        else:
            binid = int((x[i] - xmin) * nbins * norm) + 1
        count[binid] += 1
    if flow:
        count[-2] += count[-1]
        count[1] += count[0]
    return count[1:-1]


@nb.jit(
    nb.types.UniTuple(nb.float64[:], 2)(
        nb.float64[:], nb.float64[:], nb.int64, nb.float64, nb.float64, nb.boolean
    ),
    nopython=True,
)
def _float64_weighted(x, weights, nbins, xmin, xmax, flow):
    norm = 1.0 / (xmax - xmin)
    count = np.zeros((nbins + 2), dtype=np.float64)
    sumw2 = np.zeros((nbins + 2), dtype=np.float64)
    for i in range(x.shape[0]):
        binid = 0
        if x[i] < xmin:
            pass
        elif x[i] > xmax:
            binid = nbins + 1
        else:
            binid = int((x[i] - xmin) * nbins * norm) + 1
        weight = weights[i]
        count[binid] += weight
        sumw2[binid] += weight ** 2
    if flow:
        count[-2] += count[-1]
        sumw2[-2] += sumw2[-1]
        count[1] += count[0]
        sumw2[1] += sumw2[0]
    count, sumw2 = count[1:-1], sumw2[1:-1]
    return count, np.sqrt(sumw2)


@nb.jit(
    nb.float64[:](nb.float64[:], nb.int64, nb.float64, nb.float64, nb.boolean),
    nopython=True,
)
def _float64(x, nbins, xmin, xmax, flow):
    norm = 1.0 / (xmax - xmin)
    count = np.zeros((nbins + 2), dtype=np.float64)
    for i in range(x.shape[0]):
        binid = 0
        if x[i] < xmin:
            pass
        elif x[i] > xmax:
            binid = nbins + 1
        else:
            binid = int((x[i] - xmin) * nbins * norm) + 1
        count[binid] += 1
    if flow:
        count[-2] += count[-1]
        count[1] += count[0]
    return count[1:-1]


@nb.jit(
    nb.types.UniTuple(nb.float32[:, :], 2)(
        nb.float32[:], nb.float32[:, :], nb.int64, nb.float64, nb.float64, nb.boolean
    ),
    nopython=True,
    parallel=True,
)
def _float32_multiweights(x, weights, nbins, xmin, xmax, flow):
    norm = 1.0 / (xmax - xmin)
    count = np.zeros((nbins + 2, weights.shape[1]), dtype=np.float32)
    sumw2 = np.zeros((nbins + 2, weights.shape[1]), dtype=np.float32)
    for i in range(x.shape[0]):
        binid = 0
        if x[i] < xmin:
            pass
        elif x[i] > xmax:
            binid = nbins + 1
        else:
            binid = int((x[i] - xmin) * nbins * norm) + 1
        for j in range(weights.shape[1]):
            weight = weights[i][j]
            count[binid][j] += weight
            sumw2[binid][j] += weight * weight
    if flow:
        count[-2, :] += count[-1, :]
        sumw2[-2, :] += sumw2[-1, :]
        count[1, :] += count[0, :]
        sumw2[1, :] += sumw2[0, :]

    count = count[1:-1, :]
    sumw2 = sumw2[1:-1, :]
    return count, np.sqrt(sumw2)


@nb.jit(
    nb.types.UniTuple(nb.float64[:, :], 2)(
        nb.float64[:], nb.float64[:, :], nb.int64, nb.float64, nb.float64, nb.boolean
    ),
    nopython=True,
    parallel=True,
)
def _float64_multiweights(x, weights, nbins, xmin, xmax, flow):
    norm = 1.0 / (xmax - xmin)
    count = np.zeros((nbins + 2, weights.shape[1]), dtype=np.float64)
    sumw2 = np.zeros((nbins + 2, weights.shape[1]), dtype=np.float64)
    for i in range(x.shape[0]):
        binid = 0
        if x[i] < xmin:
            pass
        elif x[i] > xmax:
            binid = nbins + 1
        else:
            binid = int((x[i] - xmin) * nbins * norm) + 1
        for j in range(weights.shape[1]):
            weight = weights[i][j]
            count[binid][j] += weight
            sumw2[binid][j] += weight * weight
    if flow:
        count[-2, :] += count[-1, :]
        sumw2[-2, :] += sumw2[-1, :]
        count[1, :] += count[0, :]
        sumw2[1, :] += sumw2[0, :]

    count = count[1:-1, :]
    sumw2 = sumw2[1:-1, :]
    return count, np.sqrt(sumw2)



from setuptools import setup
from setuptools import find_packages
import os


with open("requirements.txt") as f:
    requirements = f.read().splitlines()


this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), "rb") as f:
    long_description = f.read().decode("utf-8")


setup(
    name="humba",
    version = "0.1a2",
    scripts=[],
    packages=find_packages(exclude=["tests"]),
    description="Histogramming using numba",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Doug Davis",
    author_email="ddavis@ddavis.io",
    maintainer="Doug Davis",
    maintainer_email="ddavis@ddavis.io",
    license="BSD 3-clause",
    url="https://github.com/douglasdavis/humba",
    test_suite="tests",
    python_requires=">=3.6",
    install_requires=requirements,
    tests_require=["pytest>=4.0"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
    ],
)



import humba
import numpy as np


def test64():
    x = np.random.randn(10000)
    hh, _, hedges = humba.histogram(x, bins=20, range=(-3, 3))
    hn, nedges = np.histogram(x, bins=20, range=(-3, 3))
    assert np.allclose(hh, hn)
    assert np.allclose(hedges, nedges)


def test64_weighted():
    x = np.random.randn(10000)
    w = np.random.uniform(0.5, 1.5, x.shape[0])
    hh, _, hedges = humba.histogram(x, bins=20, range=(-3, 3), weights=w)
    hn, nedges = np.histogram(x, bins=20, range=(-3, 3), weights=w)
    assert np.allclose(hh, hn)
    assert np.allclose(hedges, nedges)


def test32():
    x = np.random.randn(10000).astype(np.float32)
    hh, _, hedges = humba.histogram(x, bins=20, range=(-3, 3))
    hn, nedges = np.histogram(x, bins=20, range=(-3, 3))
    assert np.allclose(hh, hn)
    assert np.allclose(hedges, nedges)


def test32_weighted():
    x = np.random.randn(10000).astype(np.float32)
    w = np.random.uniform(0.5, 1.5, x.shape[0]).astype(np.float32)
    hh, _, hedges = humba.histogram(x, bins=20, range=(-3, 3), weights=w)
    hn, nedges = np.histogram(x, bins=20, range=(-3, 3), weights=w)
    assert np.allclose(hh, hn)
    assert np.allclose(hedges, nedges)


def test64_manyweights():
    x = np.random.randn(10000)
    w = (0.1 + np.random.rand(10000, 20)) / 2.0
    hh, _, edges = humba.histogram_mw(x, w, bins=20, range=(-3, 3))
    for i in range(hh.shape[1]):
        ihn, inedges = np.histogram(x, bins=20, range=(-3, 3), weights=w.T[i])
        assert np.allclose(hh.T[i], ihn)


def test32_manyweights():
    x = np.random.randn(10000).astype(np.float32)
    w = (0.1 + np.random.rand(10000, 20)) / 2.0
    hh, _, edges = humba.histogram_mw(x, w, bins=20, range=(-3, 3))
    for i in range(hh.shape[1]):
        ihn, inedges = np.histogram(x, bins=20, range=(-3, 3), weights=w.T[i])
        assert np.allclose(hh.T[i], ihn)
